;; Generated package description from denote-org.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "denote-org" "0.2.0" "Denote extensions for Org mode" '((emacs "28.1") (denote "4.0.0")) :commit "57bca6df0049a5a5e28b91b819b39117cba2e5c7" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/denote-org")
